import xarray as xr

from xcube.webapi.datasets.context import DatasetsContext
from xcube.server.api import Context


def get_datasets_ctx(ctx: Context) -> DatasetsContext:
    return ctx.get_api_ctx("datasets")


def get_dataset(ctx: Context, dataset_id: str | None = None) -> xr.Dataset | None:
    return get_datasets_ctx(ctx).get_dataset(dataset_id) if dataset_id else None



import altair as alt
import numpy as np
import pandas as pd
import pyproj
import shapely
import shapely.ops
import shapely.geometry
from chartlets import Component, Input, State, Output
from chartlets.components import Box, Button, CircularProgress, Select, VegaChart

from xcube.constants import CRS_CRS84
from xcube.core.geom import mask_dataset_by_geometry, normalize_geometry
from xcube.core.gridmapping import GridMapping
from xcube.webapi.viewer.contrib import Panel
from xcube.webapi.viewer.contrib import get_dataset
from xcube.server.api import Context


panel = Panel(__name__, title="2D Histogram")


# Number of bins in x and y directions.
# This results in columns of 4096 items.
# Vega Altair's maximum is 5000.
NUM_BINS_MAX = 64


@panel.layout(State("@app", "selectedDatasetId"))
def render_panel(ctx: Context, dataset_id: str | None = None) -> Component:
    dataset = get_dataset(ctx, dataset_id)

    plot = VegaChart(id="plot", chart=None, style={"paddingTop": 6})

    var_names, var_name_1, var_name_2 = get_var_select_options(dataset)

    select_var_1 = Select(
        id="select_var_1", label="Variable 1", value=var_name_1, options=var_names
    )
    select_var_2 = Select(
        id="select_var_2", label="Variable 2", value=var_name_2, options=var_names
    )

    button = Button(id="button", text="Update", style={"maxWidth": 100})

    controls = Box(
        children=[select_var_1, select_var_2, button],
        style={
            "display": "flex",
            "flexDirection": "row",
            "alignItems": "center",
            "gap": 6,
            "padding": 6,
        },
    )

    return Box(
        children=[plot, controls],
        style={
            "display": "flex",
            "flexDirection": "column",
            "alignItems": "center",
            "width": "100%",
            "height": "100%",
            "gap": 6,
            "padding": 6,
        },
    )


@panel.callback(
    State("@app", "selectedDatasetId"),
    State("@app", "selectedTimeLabel"),
    State("@app", "selectedPlaceGeometry"),
    State("select_var_1"),
    State("select_var_2"),
    Input("button", "clicked"),
    Output("plot", "chart"),
)
def update_plot(
    ctx: Context,
    dataset_id: str | None = None,
    time_label: float | None = None,
    place_geometry: str | None = None,
    var_1_name: str | None = None,
    var_2_name: str | None = None,
    _clicked: bool | None = None,  # trigger, will always be True
) -> alt.Chart | None:
    dataset = get_dataset(ctx, dataset_id)
    if dataset is None or not place_geometry or not var_1_name or not var_2_name:
        # TODO: set error message in panel UI
        print("panel disabled")
        return None

    if "time" in dataset.coords:
        if time_label:
            dataset = dataset.sel(time=pd.Timestamp(time_label[0:-1]), method="nearest")
        else:
            dataset = dataset.isel(time=-1)

    grid_mapping = GridMapping.from_dataset(dataset)
    place_geometry = normalize_geometry(place_geometry)
    if place_geometry is not None and not grid_mapping.crs.is_geographic:
        project = pyproj.Transformer.from_crs(
            CRS_CRS84, grid_mapping.crs, always_xy=True
        ).transform
        place_geometry = shapely.ops.transform(project, place_geometry)

    if (
        place_geometry is None
        or place_geometry.is_empty
        or isinstance(place_geometry, shapely.geometry.Point)
    ):
        # TODO: set error message in panel UI
        print("2-D histogram only works for geometries with a non-zero extent.")
        return

    dataset = mask_dataset_by_geometry(dataset, place_geometry)
    if dataset is None:
        # TODO: set error message in panel UI
        print("dataset is None after masking, invalid geometry?")
        return None

    var_1_data: np.ndarray = dataset[var_1_name].values.ravel()
    var_2_data: np.ndarray = dataset[var_2_name].values.ravel()
    var_1_range = [np.nanmin(var_1_data), np.nanmax(var_1_data)]
    var_2_range = [np.nanmin(var_2_data), np.nanmax(var_2_data)]
    num_bins = min(NUM_BINS_MAX, var_1_data.size)
    hist2d, x_edges, y_edges = np.histogram2d(
        var_1_data,
        var_2_data,
        bins=num_bins,
        range=np.array([var_1_range, var_2_range]),
        density=True,
    )
    # x and y 2D arrays with bin centers
    x, y = np.meshgrid(np.arange(num_bins), np.arange(num_bins))
    # z = hist2d
    z = np.where(hist2d > 0.0, hist2d, np.nan).T
    source = pd.DataFrame(
        {var_1_name: x.ravel(), var_2_name: y.ravel(), "z": z.ravel()}
    )
    # TODO: use edges or center coordinates as tick labels.
    x_centers = x_edges[0:-1] + np.diff(x_edges) / 2
    y_centers = y_edges[0:-1] + np.diff(y_edges) / 2
    # TODO: limit number of ticks on axes to, e.g., 10.
    # TODO: allow chart to be adjusted to available container (<div>) size.
    chart = (
        alt.Chart(source)
        .mark_rect()
        .encode(
            x=alt.X(
                f"{var_1_name}:O",
                # scale=alt.Scale(bins=x_centers),
            ),
            y=alt.Y(
                f"{var_2_name}:O",
                sort="descending",
                # scale=alt.Scale(bins=y_centers),
            ),
            color=alt.Color("z:Q", scale=alt.Scale(scheme="viridis"), title="Density"),
        )
    ).properties(width=360, height=360)

    return chart


@panel.callback(
    Input("@app", "selectedDatasetId"),
    Input("@app", "selectedPlaceGeometry"),
    Output("button", "disabled"),
)
def set_button_disablement(
    _ctx: Context,
    dataset_id: str | None = None,
    place_geometry: str | None = None,
) -> bool:
    return not dataset_id or not place_geometry


@panel.callback(
    Input("@app", "selectedDatasetId"),
    State("select_var_1", "value"),
    State("select_var_2", "value"),
    Output("select_var_1", "options"),
    Output("select_var_1", "value"),
    Output("select_var_2", "options"),
    Output("select_var_2", "value"),
)
def populate_selects(
    ctx: Context,
    dataset_id: str | None = None,
    var_name_1: str | None = None,
    var_name_2: str | None = None,
) -> tuple[list, str | None, list, str | None]:
    dataset = get_dataset(ctx, dataset_id)
    var_names, var_name_1, var_name_2 = get_var_select_options(
        dataset, var_name_1, var_name_2
    )
    return var_names, var_name_1, var_names, var_name_2


def get_var_select_options(
    dataset,
    var_name_1: str | None = None,
    var_name_2: str | None = None,
) -> tuple[list, str | None, str | None]:

    if dataset is not None:
        var_names = [
            var_name
            for var_name, var in dataset.data_vars.items()
            if len(var.dims) >= 1
        ]
    else:
        var_names = []

    if var_names:
        if not var_name_1 or var_name_1 not in var_names:
            var_name_1 = var_names[0]
        if not var_name_2 or var_name_2 not in var_names:
            var_name_2 = var_names[0]

    return var_names, var_name_1, var_name_2


# TODO: Doesn't work. We need to ensure that show_progress() returns
#   before update_plot()
# @panel.callback(
#     Input("button", "clicked"),
#     Output("button", ""),
# )
def show_progress(
    _ctx: Context,
    _clicked: bool | None = None,  # trigger, will always be True
) -> alt.Chart | None:
    return CircularProgress(id="button", size=28)



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.util.jsonschema import JsonObjectSchema

# Note: we could limit/enhance the default expression capabilities here
CONFIG_SCHEMA = JsonObjectSchema(additional_properties=True)



from chartlets import Component, Input, State, Output
from chartlets.components import Box, Select, Checkbox, Typography

from xcube.webapi.viewer.contrib import Panel
from xcube.webapi.viewer.contrib import get_datasets_ctx
from xcube.server.api import Context


panel = Panel(__name__, title="Panel B")


COLORS = [(0, "red"), (1, "green"), (2, "blue"), (3, "yellow")]


@panel.layout(
    State("@app", "selectedDatasetId"),
)
def render_panel(
    ctx: Context,
    dataset_id: str = "",
) -> Component:

    opaque = False
    color = 0

    opaque_checkbox = Checkbox(
        id="opaque",
        value=opaque,
        label="Opaque",
    )

    color_select = Select(
        id="color",
        value=color,
        label="Color",
        options=COLORS,
        style={"flexGrow": 0, "minWidth": 80},
    )

    info_text = Typography(
        id="info_text", children=update_info_text(ctx, dataset_id, opaque, color)
    )

    return Box(
        style={
            "display": "flex",
            "flexDirection": "column",
            "width": "100%",
            "height": "100%",
            "gap": "6px",
        },
        children=[opaque_checkbox, color_select, info_text],
    )


# noinspection PyUnusedLocal
@panel.callback(
    Input("@app", "selectedDatasetId"),
    Input("opaque"),
    Input("color"),
    State("info_text", "children"),
    Output("info_text", "children"),
)
def update_info_text(
    ctx: Context,
    dataset_id: str = "",
    opaque: bool = False,
    color: int = 0,
    info_children: list[str] = "",
) -> list[str]:
    ds_ctx = get_datasets_ctx(ctx)
    ds_configs = ds_ctx.get_dataset_configs()

    info_text = info_children[0] if info_children else ""

    opaque = opaque or False
    color = color if color is not None else 0
    return [
        f"The dataset is {dataset_id},"
        f" the color is {COLORS[color][1]} and"
        f" it {'is' if opaque else 'is not'} opaque."
        f" The length of the last info text"
        f" was {len(info_text or "")}."
        f" The number of datasets is {len(ds_configs)}."
    ]



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from contextlib import contextmanager
from functools import cached_property
from pathlib import Path
from typing import Optional, Any
from collections.abc import Mapping, MutableMapping
import sys

from chartlets import Extension
from chartlets import ExtensionContext
import fsspec

from xcube.constants import LOG
from xcube.server.api import Context
from xcube.util.fspath import is_local_fs
from xcube.util.temp import new_temp_dir
from xcube.webapi.common.context import ResourcesContext
from xcube.webapi.viewer.contrib import Panel

Extension.add_contrib_point("panels", Panel)


class ViewerContext(ResourcesContext):
    """Context for xcube Viewer API."""

    def __init__(self, server_ctx: Context):
        super().__init__(server_ctx)
        self.ext_ctx: ExtensionContext | None = None
        self.persistence: MutableMapping | None = None

    def on_update(self, prev_context: Optional[Context]):
        super().on_update(prev_context)
        viewer_config: dict = self.config.get("Viewer")
        if not viewer_config:
            return
        persistence: dict | None = viewer_config.get("Persistence")
        if persistence:
            path = self.get_config_path(persistence, "Persistence")
            storage_options = persistence.get("StorageOptions")
            self.set_persistence(path, storage_options)
        augmentation: dict | None = viewer_config.get("Augmentation")
        if augmentation:
            path = augmentation.get("Path")
            extension_refs = augmentation["Extensions"]
            self.set_extension_context(path, extension_refs)

    @cached_property
    def config_items(self) -> Optional[Mapping[str, bytes]]:
        if self.config_path is None:
            return None
        return fsspec.get_mapper(self.config_path)

    @cached_property
    def config_path(self) -> Optional[str]:
        if "Viewer" not in self.config:
            return None
        return self.get_config_path(
            self.config["Viewer"].get("Configuration", {}),
            "'Configuration' item of 'Viewer'",
        )

    def set_persistence(self, path: str, storage_options: dict[str, Any] | None):
        fs_root: tuple[fsspec.AbstractFileSystem, str] = fsspec.core.url_to_fs(
            path, **(storage_options or {})
        )
        fs, root = fs_root
        self.persistence = fs.get_mapper(root, create=True, check=True)
        LOG.info(f"Viewer persistence established for path {path!r}")

    def set_extension_context(self, path: str | None, extension_refs: list[str]):
        module_path = self.base_dir
        if path:
            module_path = f"{module_path}/{path}"
        fs_root: tuple[fsspec.AbstractFileSystem, str] = fsspec.core.url_to_fs(
            module_path
        )
        fs, fs_path = fs_root
        if is_local_fs(fs):
            local_module_path = Path(module_path)
        else:
            temp_module_path = new_temp_dir("xcube-viewer-aux-")
            LOG.warning(f"Downloading {module_path!r} to {temp_module_path!r}")
            fs.get(fs_path + "/**/*", temp_module_path + "/", recursive=True)
            local_module_path = Path(temp_module_path)
        with prepend_sys_path(local_module_path):
            LOG.info(f"Loading viewer extension(s) {','.join(extension_refs)}")
            self.ext_ctx = ExtensionContext.load(self.server_ctx, extension_refs)


@contextmanager
def prepend_sys_path(path: Path | str | None):
    prev_sys_path = None
    if path is not None:
        LOG.warning(f"Temporarily prepending '{path}' to sys.path")
        prev_sys_path = sys.path
        sys.path = [str(path)] + sys.path
    try:
        yield path is not None
    finally:
        if prev_sys_path:
            sys.path = prev_sys_path
            LOG.info(f"Restored sys.path")



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Optional, Dict, Union, List, Any

from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonNumberSchema
from xcube.util.jsonschema import JsonObject
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema
from ..response import CubeGeneratorResult
from ..response import CubeInfo
from ..response import make_cube_generator_result_class


class CubeGeneratorToken(JsonObject):
    def __init__(self, access_token: str, token_type: str):
        self.access_token = access_token
        self.token_type = token_type

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                access_token=JsonStringSchema(min_length=1),
                token_type=JsonStringSchema(min_length=1),
            ),
            required=["access_token", "token_type"],
            additional_properties=False,
            factory=cls,
        )

    @classmethod
    def from_dict(cls, value: dict) -> "CubeGeneratorToken":
        return cls.get_schema().from_instance(value)


class CubeGeneratorJobStatus(JsonObject):
    # noinspection PyUnusedLocal
    def __init__(
        self,
        succeeded: int = None,
        failed: int = None,
        active: int = None,
        start_time: str = None,
        completion_time: str = None,
        conditions: list[dict[str, Any]] = None,
        **additional_properties
    ):
        self.succeeded: Optional[int] = succeeded
        self.failed: Optional[int] = failed
        self.active: Optional[int] = active
        self.start_time: Optional[str] = start_time
        self.completion_time: Optional[str] = completion_time
        self.conditions: Optional[dict[str, Any]] = conditions
        self.additional_properties: dict[str, Any] = additional_properties

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                succeeded=JsonIntegerSchema(nullable=True),
                failed=JsonIntegerSchema(nullable=True),
                active=JsonIntegerSchema(nullable=True),
                start_time=JsonStringSchema(nullable=True),
                completion_time=JsonStringSchema(nullable=True),
                conditions=JsonArraySchema(
                    items=JsonObjectSchema(additional_properties=True), nullable=True
                ),
            ),
            additional_properties=True,
            factory=cls,
        )

    @classmethod
    def from_dict(cls, value: dict) -> "CubeGeneratorJobStatus":
        return cls.get_schema().from_instance(value)


class CubeGeneratorProgressState(JsonObject):
    """Current progress state of the remote generator."""

    def __init__(
        self,
        progress: float,
        # worked: Union[int, float],
        total_work: Union[int, float],
        **additional_properties
    ):
        self.progress: float = float(progress)
        # self.worked: float = float(worked)
        self.total_work: float = float(total_work)
        self.additional_properties = additional_properties

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                progress=JsonNumberSchema(minimum=0.0),
                # worked=JsonNumberSchema(minimum=0),
                total_work=JsonNumberSchema(exclusive_minimum=0),
            ),
            required=[
                "progress",
                # 'worked',
                "total_work",
            ],
            additional_properties=True,
            factory=cls,
        )

    @classmethod
    def from_dict(cls, value: dict) -> "CubeGeneratorProgressState":
        return cls.get_schema().from_instance(value)


class CubeGeneratorProgress(JsonObject):
    """Current progress of the remote generator."""

    def __init__(self, sender: str, state: CubeGeneratorProgressState):
        self.sender: str = sender
        self.state: CubeGeneratorProgressState = state

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                sender=JsonStringSchema(),
                state=CubeGeneratorProgressState.get_schema(),
            ),
            required=[
                "sender",
                "state",
            ],
            additional_properties=False,
            factory=cls,
        )

    @classmethod
    def from_dict(cls, value: dict) -> "CubeGeneratorProgress":
        return cls.get_schema().from_instance(value)


class CubeGeneratorState(JsonObject):
    """Current state of the remote generator."""

    def __init__(
        self,
        job_id: str,
        job_status: CubeGeneratorJobStatus,
        job_result: Optional[CubeGeneratorResult] = None,
        output: Optional[list[str]] = None,
        progress: Optional[list[CubeGeneratorProgress]] = None,
        **additional_properties
    ):
        self.job_id = job_id
        self.job_status = job_status
        self.job_result = job_result
        self.output = output
        self.progress = progress
        self.additional_properties = additional_properties

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                job_id=JsonStringSchema(min_length=1),
                job_status=CubeGeneratorJobStatus.get_schema(),
                job_result=CubeGeneratorResult.get_schema(),
                output=JsonArraySchema(items=JsonStringSchema(), nullable=True),
                progress=JsonArraySchema(
                    items=CubeGeneratorProgress.get_schema(), nullable=True
                ),
            ),
            required=["job_id", "job_status"],
            additional_properties=True,
            factory=cls,
        )

    @classmethod
    def from_dict(cls, value: dict) -> "CubeGeneratorState":
        return cls.get_schema().from_instance(value)


class CostEstimation(JsonObject):
    def __init__(self, required: int, available: int = None, limit: int = None):
        self.available: int = available
        self.limit: Optional[int] = limit
        self.required: Optional[int] = required

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                required=JsonIntegerSchema(minimum=0),
                available=JsonIntegerSchema(minimum=0),
                limit=JsonIntegerSchema(minimum=0),
            ),
            required=["required"],
            additional_properties=False,
            factory=cls,
        )

    @classmethod
    def from_dict(cls, value: dict) -> "CostEstimation":
        return cls.get_schema().from_instance(value)


class CubeInfoWithCosts(CubeInfo):
    def __init__(self, cost_estimation: CostEstimation, **kwargs):
        super().__init__(**kwargs)
        self.cost_estimation: CostEstimation = cost_estimation

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        schema = super().get_schema()
        schema.properties.update(cost_estimation=CostEstimation.get_schema())
        schema.required.append("cost_estimation")
        schema.factory = cls
        return schema


CubeInfoWithCostsResult = make_cube_generator_result_class(CubeInfoWithCosts)



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import collections.abc
import numbers
from typing import Optional, Any, Tuple, Union
from collections.abc import Sequence, Mapping, Iterable

import pyproj

from xcube.util.assertions import assert_given
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonBooleanSchema
from xcube.util.jsonschema import JsonDateSchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonNumberSchema
from xcube.util.jsonschema import JsonObject
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema


class InputConfig(JsonObject):
    def __init__(
        self,
        store_id: str = None,
        opener_id: str = None,
        data_id: str = None,
        store_params: Mapping[str, Any] = None,
        open_params: Mapping[str, Any] = None,
    ):
        assert_true(
            store_id or opener_id, "One of store_id and opener_id must be given"
        )
        assert_given(data_id, "data_id")
        self.store_id = store_id
        self.opener_id = opener_id
        self.data_id = data_id
        self.store_params = store_params
        self.open_params = open_params

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                store_id=JsonStringSchema(min_length=1),
                opener_id=JsonStringSchema(min_length=1),
                data_id=JsonStringSchema(min_length=1),
                store_params=JsonObjectSchema(
                    additional_properties=True, nullable=True
                ),
                open_params=JsonObjectSchema(additional_properties=True, nullable=True),
            ),
            additional_properties=False,
            required=["data_id"],
            factory=cls,
        )


class CallbackConfig(JsonObject):
    def __init__(self, api_uri: str = None, access_token: str = None):
        assert_true(
            api_uri and access_token, "Both, api_uri and access_token must be given"
        )
        self.api_uri = api_uri
        self.access_token = access_token

    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(
                api_uri=JsonStringSchema(min_length=1),
                access_token=JsonStringSchema(min_length=1),
            ),
            additional_properties=False,
            required=["api_uri", "access_token"],
            factory=cls,
        )


class OutputConfig(JsonObject):
    def __init__(
        self,
        store_id: str = None,
        writer_id: str = None,
        data_id: str = None,
        store_params: Mapping[str, Any] = None,
        write_params: Mapping[str, Any] = None,
        replace: bool = None,
    ):
        assert_true(
            store_id or writer_id, "One of store_id and writer_id must be given"
        )
        self.store_id = store_id
        self.writer_id = writer_id
        self.data_id = data_id
        self.store_params = store_params
        self.write_params = write_params
        self.replace = replace

    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(
                store_id=JsonStringSchema(min_length=1),
                writer_id=JsonStringSchema(min_length=1),
                data_id=JsonStringSchema(default=None),
                store_params=JsonObjectSchema(
                    additional_properties=True, nullable=True
                ),
                write_params=JsonObjectSchema(
                    additional_properties=True, nullable=True
                ),
                replace=JsonBooleanSchema(default=False),
            ),
            additional_properties=False,
            required=[],
            factory=cls,
        )


# Need to be aligned with params in resample_cube(cube, **params)
class CubeConfig(JsonObject):
    def __init__(
        self,
        variable_names: Sequence[str] = None,
        crs: str = None,
        bbox: tuple[float, float, float, float] = None,
        spatial_res: Union[float, tuple[float]] = None,
        tile_size: Union[int, tuple[int, int]] = None,
        time_range: tuple[str, Optional[str]] = None,
        time_period: str = None,
        chunks: Mapping[str, Optional[int]] = None,
        metadata: Mapping[str, Any] = None,
        variable_metadata: Mapping[str, Mapping[str, Any]] = None,
    ):
        self.variable_names = None
        if variable_names is not None:
            assert_true(len(variable_names) > 0, "variable_names is invalid")
            self.variable_names = tuple(map(str, variable_names))

        self.crs = None
        if crs is not None:
            assert_instance(crs, str, "crs")
            try:
                pyproj.crs.CRS.from_string(crs)
            except pyproj.exceptions.CRSError:
                raise ValueError("crs is invalid")
            self.crs = crs

        self.bbox = None
        if bbox is not None:
            assert_true(len(bbox) == 4, "bbox is invalid")
            self.bbox = tuple(map(float, bbox))

        self.spatial_res = None
        if spatial_res is not None:
            assert_instance(spatial_res, numbers.Number, "spatial_res")
            self.spatial_res = float(spatial_res)

        self.tile_size = None
        if tile_size is not None:
            if isinstance(tile_size, int):
                tile_width, tile_height = tile_size, tile_size
            else:
                try:
                    tile_width, tile_height = tile_size
                except (ValueError, TypeError):
                    raise ValueError(
                        "tile_size must be an integer " "or a pair of integers"
                    )
                assert_instance(tile_width, numbers.Number, "tile_width of tile_size")
                assert_instance(tile_height, numbers.Number, "tile_height of tile_size")
            self.tile_size = tile_width, tile_height

        self.time_range = None
        if time_range is not None:
            assert_true(len(time_range) == 2, "time_range is invalid")
            self.time_range = tuple(time_range)

        self.time_period = None
        if time_period is not None:
            assert_instance(time_period, str, "time_period")
            self.time_period = time_period

        self.chunks = None
        if chunks is not None:
            assert_instance(chunks, collections.abc.Mapping, "chunks")
            for chunk_dim, chunk_size in chunks.items():
                assert_instance(chunk_dim, str, "chunk dimension")
                assert_instance(chunk_size, (int, type(None)), "chunk size")
            self.chunks = dict(chunks)

        self.metadata = None
        if metadata is not None:
            assert_instance(metadata, collections.abc.Mapping, "metadata")
            self.metadata = dict(metadata)

        self.variable_metadata = None
        if variable_metadata is not None:
            assert_instance(
                variable_metadata, collections.abc.Mapping, "variable_metadata"
            )
            for var_name, metadata in variable_metadata.items():
                assert_instance(var_name, str, "chunk dimension")
                assert_instance(metadata, collections.abc.Mapping, "variable metadata")
            self.variable_metadata = dict(variable_metadata)

    def drop_props(self, name: Union[str, Iterable[str]]):
        """Drop one or more named properties from this configuration.

        Args:
            name: the name of the property to be dropped.
            names: more names of properties to be dropped.

        Returns:
            a new cube configuration.
        """
        if not name:
            return self
        name_set = {name} if isinstance(name, str) else set(name)
        for k in name_set:
            assert_true(
                hasattr(self, k), message=f"{k} is not a property of {CubeConfig!r}"
            )
        return self.from_dict(
            {k: v for k, v in self.to_dict().items() if k not in name_set}
        )

    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(
                variable_names=JsonArraySchema(
                    nullable=True, items=JsonStringSchema(min_length=1), min_items=0
                ),
                crs=JsonStringSchema(nullable=True, min_length=1),
                bbox=JsonArraySchema(
                    nullable=True,
                    items=[
                        JsonNumberSchema(),
                        JsonNumberSchema(),
                        JsonNumberSchema(),
                        JsonNumberSchema(),
                    ],
                ),
                spatial_res=JsonNumberSchema(nullable=True, exclusive_minimum=0.0),
                tile_size=JsonArraySchema(
                    nullable=True,
                    items=[
                        JsonIntegerSchema(minimum=1, maximum=2500),
                        JsonIntegerSchema(minimum=1, maximum=2500),
                    ],
                ),
                time_range=JsonDateSchema.new_range(nullable=True),
                time_period=JsonStringSchema(
                    nullable=True, pattern=r"^([1-9][0-9]*)?[DWMY]$"
                ),
                chunks=JsonObjectSchema(
                    nullable=True,
                    additional_properties=JsonIntegerSchema(nullable=True, minimum=1),
                ),
                metadata=JsonObjectSchema(nullable=True, additional_properties=True),
                variable_metadata=JsonObjectSchema(
                    nullable=True,
                    additional_properties=JsonObjectSchema(additional_properties=True),
                ),
            ),
            additional_properties=False,
            factory=cls,
        )



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.


from .abc import MultiLevelDataset
from .base import BaseMultiLevelDataset
from .combined import CombinedMultiLevelDataset
from .computed import ComputedMultiLevelDataset
from .computed import augment_ml_dataset
from .computed import open_ml_dataset_from_python_code
from .fs import FsMultiLevelDataset
from .fs import FsMultiLevelDatasetError
from .identity import IdentityMultiLevelDataset
from .lazy import LazyMultiLevelDataset
from .mapped import MappedMultiLevelDataset



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.


from xcube.server.api import Context
from xcube.webapi.common.context import ResourcesContext
from ...datasets.context import DatasetsContext
from ...tiles.context import TilesContext


class WmtsContext(ResourcesContext):
    _feature_index: int = 0

    def __init__(self, server_ctx: Context):
        super().__init__(server_ctx)
        self._tiles_ctx = server_ctx.get_api_ctx("tiles", cls=TilesContext)
        self._datasets_ctx = server_ctx.get_api_ctx("datasets", cls=DatasetsContext)

    @property
    def tiles_ctx(self) -> TilesContext:
        return self._tiles_ctx

    @property
    def datasets_ctx(self) -> DatasetsContext:
        return self._datasets_ctx



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .accessor import DataDeleter
from .accessor import DataOpener
from .accessor import DataPreloader
from .accessor import DataTimeSliceUpdater
from .accessor import DataWriter
from .accessor import find_data_opener_extensions
from .accessor import find_data_writer_extensions
from .accessor import get_data_accessor_predicate
from .accessor import new_data_opener
from .accessor import new_data_writer
from .assertions import assert_valid_config
from .assertions import assert_valid_params
from .datatype import ANY_TYPE
from .datatype import DATASET_TYPE
from .datatype import DataType
from .datatype import DataTypeLike
from .datatype import GEO_DATA_FRAME_TYPE
from .datatype import MULTI_LEVEL_DATASET_TYPE
from .descriptor import DataDescriptor
from .descriptor import DatasetDescriptor
from .descriptor import GeoDataFrameDescriptor
from .descriptor import MultiLevelDatasetDescriptor
from .descriptor import VariableDescriptor
from .descriptor import new_data_descriptor
from .error import DataStoreError
from .fs.registry import new_fs_data_store
from .preload import PreloadHandle
from .preload import PreloadState
from .preload import PreloadStatus
from .search import DataSearcher
from .search import DefaultSearchMixin
from .store import DataStore
from .store import MutableDataStore
from .store import find_data_store_extensions
from .store import list_data_store_ids
from .store import get_data_store_class
from .store import get_data_store_params_schema
from .store import new_data_store
from .storepool import DataStoreConfig
from .storepool import DataStoreInstance
from .storepool import DataStorePool
from .storepool import DataStorePoolLike
from .storepool import get_data_store_instance



from chartlets import Contribution


class Panel(Contribution):
    """Panel contribution.

    A panel is a UI-contribution to xcube Viewer.
    To become effective, instances of this class must be added
    to a ``chartlets.Extension`` instance exported from your extension
    module.

    Args:
        name: A name that is unique within the extension.
        title: An initial title for the panel.
    """

    def __init__(self, name: str, title: str | None = None):
        super().__init__(name, visible=False, title=title)



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .routes import PATH_PARAM_DATASET_ID
from .routes import PATH_PARAM_VAR_NAME
from .routes import QUERY_PARAM_CMAP
from .routes import QUERY_PARAM_NORM
from .routes import QUERY_PARAM_CRS
from .routes import QUERY_PARAM_VMAX
from .routes import QUERY_PARAM_VMIN
from .routes import api



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import List, Optional, Any
from collections.abc import Mapping

from xcube.constants import LOG
from xcube.server.api import ApiError
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonBooleanSchema
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema

AUTHENTICATION_SCHEMA = JsonObjectSchema(
    properties=dict(
        Authority=JsonStringSchema(format="uri"),
        Domain=JsonStringSchema(description="Deprecated, use Authority"),
        Audience=JsonStringSchema(),
        Algorithms=JsonArraySchema(
            items=JsonStringSchema(default=["RS256"]),
        ),
        IsRequired=JsonBooleanSchema(default=False),
    ),
    required=[
        # Either "Authority" or "Domain" are required
        # "Authority",
        "Audience"
    ],
    additional_properties=False,
)


class AuthConfig:
    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(Authentication=AUTHENTICATION_SCHEMA),
            factory=AuthConfig.from_dict,
        )

    def __init__(
        self,
        authority: str,
        audience: str,
        algorithms: list[str],
        is_required: bool = False,
    ):
        self._authority = authority
        self._audience = audience
        self._algorithms = algorithms
        self._is_required = is_required

    @property
    def authority(self) -> str:
        return self._authority

    @property
    def _norm_authority(self) -> str:
        return self.authority[:-1] if self.authority.endswith("/") else self.authority

    @property
    def well_known_oid_config(self) -> str:
        return f"{self._norm_authority}/.well-known/openid-configuration"

    @property
    def well_known_jwks(self) -> str:
        return f"{self._norm_authority}/.well-known/jwks.json"

    @property
    def audience(self) -> str:
        return self._audience

    @property
    def algorithms(self) -> list[str]:
        return self._algorithms

    @property
    def is_required(self) -> bool:
        return self._is_required

    @classmethod
    def from_config(cls, config: Mapping[str, Any]) -> Optional["AuthConfig"]:
        authentication = config.get("Authentication")
        if authentication is None:
            return None
        return cls.from_dict(authentication)

    @classmethod
    def from_dict(cls, config: Mapping[str, Any]) -> Optional["AuthConfig"]:
        authority = config.get("Authority")
        domain = config.get("Domain")
        if domain:
            LOG.warning(
                'Configuration parameter "Domain"'
                ' in section "Authentication"'
                ' has been deprecated. Use "Authority" instead.'
            )
            if not authority:
                authority = f"https://{domain}"
        if not authority:
            raise ApiError.InvalidServerConfig(
                'Missing key "Authority" in section "Authentication"'
            )
        audience = config.get("Audience")
        if not audience:
            raise ApiError.InvalidServerConfig(
                'Missing key "Audience" in section "Authentication"'
            )
        algorithms = config.get("Algorithms", ["RS256"])
        if not algorithms:
            raise ApiError.InvalidServerConfig(
                'Value for key "Algorithms" in section'
                ' "Authentication" must not be empty'
            )
        is_required = config.get("IsRequired", False)
        return AuthConfig(authority, audience, algorithms, is_required=is_required)



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

# noinspection PyUnresolvedReferences
from xcube.core.byoa import CodeConfig

# noinspection PyUnresolvedReferences
from xcube.core.byoa import FileSet

# noinspection PyUnresolvedReferences
from xcube.core.store.descriptor import DatasetDescriptor

from .config import CallbackConfig
from .config import CubeConfig
from .config import InputConfig
from .config import OutputConfig
from .error import CubeGeneratorError
from .generator import CubeGenerator
from .local.generator import LocalCubeGenerator
from .processor import DatasetProcessor
from .processor import METHOD_NAME_DATASET_PROCESSOR
from .processor import METHOD_NAME_PARAMS_SCHEMA_GETTER
from .remote.config import ServiceConfig
from .remote.config import ServiceConfigLike
from .remote.generator import RemoteCubeGenerator
from .remote.response import CostEstimation
from .remote.response import CubeInfoWithCosts
from .remote.response import CubeInfoWithCostsResult
from .request import CubeGeneratorRequest
from .request import CubeGeneratorRequestLike
from .response import CubeGeneratorResult
from .response import CubeInfo
from .response import CubeInfoResult
from .response import CubeReference



# Copyright (c) 2018-2024 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .routes import PATH_PARAM_X
from .routes import PATH_PARAM_Y
from .routes import PATH_PARAM_Z
from .routes import QUERY_PARAM_TIME
from .routes import QUERY_PARAM_FORMAT
from .routes import QUERY_PARAM_RETINA
from .routes import api
